import numpy as np
import scipy.stats as sps

from .analytic import Analytic

import marshmallow as m
import logging
logger = logging.getLogger(__name__)


class Beta(Analytic):
    class ConfigSchema(Analytic.ConfigSchema):
        class Meta:
            include = {
                'a': m.fields.Float(missing=1),
                'b': m.fields.Float(missing=1),
            }

    def load_config(self, config):
        super().load_config(config)
        self.rv = sps.beta(
            a=config['a'],
            b=config['b'],
            loc=0,
            scale=1)

    def get_x_min(self):
        return 0.0

    def get_x_max(self):
        return 1.0



import unittest
import math

from . import *

import logging
logger = logging.getLogger(__name__)


class TestConstant(unittest.TestCase):
    def test_v(self):
        test_var = Constant()
        self.assertEqual(test_var.get_random_value(), 0)

        constant_var0 = Constant({'loc': 0})
        self.assertEqual(constant_var0.get_random_value(), 0)

        constant_var1 = Constant({'loc': 1})
        self.assertEqual(constant_var1.get_random_value(), 1)

    def test_cdf(self):
        constant_var10 = Constant({'loc': 10})
        self.assertEqual(constant_var10.cdf(20), 200)
        self.assertEqual(constant_var10.pdf(0), 10)
        self.assertEqual(constant_var10.pdf(1), 10)


class TestUniform(unittest.TestCase):
    def test_uniform(self):
        test_var = Uniform({'max': 2.0})
        cdf = test_var.cdf(0.5)
        self.assertAlmostEqual(cdf, 0.25)
        pdf = test_var.pdf(0.5)
        self.assertAlmostEqual(pdf, 0.5)


class TestLogistic(unittest.TestCase):
    def test_cdf(self):
        test_var = Logistic()
        cdf0 = test_var.cdf(0)
        self.assertEqual(cdf0, 0.5)


class TestNormal(unittest.TestCase):
    def test_cdf(self):
        test_var = Normal()
        cdf = test_var.cdf(0.0)
        self.assertAlmostEqual(cdf, 0.5)

        cdf = test_var.cdf(1.0)
        self.assertAlmostEqual(cdf, 0.841344746)

    def test_pdf(self):
        test_var = Normal()
        self.assertAlmostEqual(test_var.pdf(2.0), 0.05399096651)


class TestSkewNormal(unittest.TestCase):
    def test_cdf(self):
        test_var = SkewNormal({'a': 1})
        cdf = test_var.cdf(0)
        self.assertAlmostEqual(cdf, 0.25)

        test_var = SkewNormal({'a': -1})
        cdf = test_var.cdf(0)
        self.assertAlmostEqual(cdf, 0.75)


class TestLogNormal(unittest.TestCase):
    def test_cdf(self):
        test_var = LogNormal()

        p50 = test_var.get_percentile(.5)
        self.assertAlmostEqual(p50, 1.0)

        p75 = test_var.get_percentile(0.75)
        cdf = test_var.cdf(p75)
        self.assertAlmostEqual(cdf, 0.75)

    def test_pdf(self):
        test_var = LogNormal({'loc':0, 'scale':1})
        self.assertAlmostEqual(test_var.pdf(1.5), 0.24497365)


class TestGamma(unittest.TestCase):
    def test_gamma1(self):
        test_var = Gamma({'loc': 1})
        pdf0 = test_var.pdf(1)
        self.assertAlmostEqual(pdf0, 1.0)

    def test_gamma_half(self):
        test_var = Gamma({'rate': 0.5})
        pdf1 = test_var.pdf(1)
        self.assertAlmostEqual(pdf1, 0.30326532)


class TestBeta(unittest.TestCase):
    def test_pdf(self):
        test_var = Beta({})
        pdf = test_var.pdf(0.1)
        self.assertAlmostEqual(pdf, 1.0)



import numpy as np
import scipy.stats as sps

from .analytic import Analytic

import marshmallow as m
import logging
logger = logging.getLogger(__name__)


class SkewNormal(Analytic):
    class ConfigSchema(Analytic.ConfigSchema):
        loc = m.fields.Float(missing=0)
        scale = m.fields.Float(missing=1)
        a = m.fields.Float(missing=0)

    def load_config(self, config):
        super().load_config(config)
        self.rv = sps.skewnorm(
        	loc=config['loc'],
        	scale=config['scale'],
        	a=config['a'])



import unittest
import math
import json

from pydantic import BaseModel, parse_obj_as

from . import *

import logging
logger = logging.getLogger(__name__)


# test_list = [
#     {
#         'dist': {
#             't': 'logistic',
#             'loc': 1
#         },
#         'cdf': {'x': 1, 'y': 0.5}
#     }
# ]
#parse_obj_as(Distribution, config)


class TestConstant(unittest.TestCase):
    def test_v(self):
        test_var = factory.get_instance({'t': 'constant'})
        self.assertEqual(test_var.get_random_value(), 0)

        test_var0 = factory.get_instance({'t': 'constant', 'loc': 0})
        self.assertEqual(test_var0.get_random_value(), 0)

        test_var1 = factory.get_instance({'t': 'constant', 'loc': 1})
        self.assertEqual(test_var1.get_random_value(), 1)

    def test_cdf(self):
        test_var = factory.get_instance({'t': 'constant', 'loc': 10})
        self.assertEqual(test_var.cdf(20), 200)
        self.assertEqual(test_var.pdf(0), 10)
        self.assertEqual(test_var.pdf(1), 10)


class TestUniform(unittest.TestCase):
    def test_uniform(self):
        test_var = factory.get_instance({'t': 'uniform', 'b': 2})
        cdf = test_var.cdf(0.5)
        self.assertAlmostEqual(cdf, 0.25)
        pdf = test_var.pdf(0.5)
        self.assertAlmostEqual(pdf, 0.5)


class TestLogistic(unittest.TestCase):
    def test_cdf(self):
        test_var = factory.get_instance({'t': 'logistic', 'loc': 1})
        #logger.debug(json.dumps(test_var.dict(), indent=2))

        cdf0 = test_var.cdf(1)
        self.assertEqual(cdf0, 0.5)


class TestNormal(unittest.TestCase):
    def test_cdf(self):
        test_var = factory.get_instance({'t': 'normal'})
        cdf = test_var.cdf(0.0)
        self.assertAlmostEqual(cdf, 0.5)

        cdf = test_var.cdf(1.0)
        self.assertAlmostEqual(cdf, 0.841344746)

    def test_pdf(self):
        test_var = factory.get_instance({'t': 'normal'})
        self.assertAlmostEqual(test_var.pdf(2.0), 0.05399096651)


class TestSkewNormal(unittest.TestCase):
    def test_cdf(self):
        test_var = factory.get_instance({'t': 'skew_normal', 'a': 1})
        cdf = test_var.cdf(0)
        self.assertAlmostEqual(cdf, 0.25)

        test_var = factory.get_instance({'t': 'skew_normal', 'a': -1})
        cdf = test_var.cdf(0)
        self.assertAlmostEqual(cdf, 0.75)


class TestLogNormal(unittest.TestCase):
    def test_cdf(self):
        test_var = factory.get_instance({'t': 'log_normal'})

        p50 = test_var.get_percentile(.5)
        self.assertAlmostEqual(p50, 1.0)

        p75 = test_var.get_percentile(0.75)
        cdf = test_var.cdf(p75)
        self.assertAlmostEqual(cdf, 0.75)

    def test_pdf(self):
        test_var = factory.get_instance({'t': 'log_normal', 'loc': 0, 'scale': 1})
        self.assertAlmostEqual(test_var.pdf(1.5), 0.24497365)


class TestGamma(unittest.TestCase):
    def test_gamma1(self):
        test_var = factory.get_instance({'t': 'gamma', 'loc': 1})
        pdf0 = test_var.pdf(1)
        self.assertAlmostEqual(pdf0, 1.0)

    def test_gamma_half(self):
        test_var = factory.get_instance({'t': 'gamma', 'rate': 0.5})
        pdf1 = test_var.pdf(1)
        self.assertAlmostEqual(pdf1, 0.30326532)


class TestBeta(unittest.TestCase):
    def test_pdf(self):
        test_var = factory.get_instance({'t': 'beta'})
        pdf = test_var.pdf(0.1)
        self.assertAlmostEqual(pdf, 1.0)



from pydantic import BaseModel, Field
from typing import Union
from typing_extensions import Annotated

from keyrock_model.model import Factory

from .beta import Beta
from .constant import Constant
from .gamma import Gamma
from .logistic import Logistic
from .log_normal import LogNormal
from .normal import Normal
from .skew_normal import SkewNormal
from .uniform import Uniform

Distribution = Annotated[
                Union[
                    Beta,
                    Constant,
                    Gamma,
                    Logistic,
                    LogNormal,
                    Normal,
                    SkewNormal,
                    Uniform,
                ],
                Field(discriminator='t')
            ]

factory = Factory(Distribution)



import numpy as np
import scipy.stats as sps
from scipy.special import erfinv
import numpy as np
import math

from .analytic import Analytic

import marshmallow as m
import logging
logger = logging.getLogger(__name__)


SQRT2 = math.sqrt(2)
SQRT2PI = math.sqrt(2 * math.pi)
MIN_X = 0.000000001

class LogNormal(Analytic):
    class ConfigSchema(Analytic.ConfigSchema):
        loc = m.fields.Float(missing=0)
        scale = m.fields.Float(missing=1)

    def load_config(self, config):
        super().load_config(config)
        self.mu = config['loc']
        self.sigma = config['scale']

        # a = 1 + (sigma / mu) ** 2
        # s = np.sqrt(np.log(a))
        # scale = mu / np.sqrt(a)
        # logger.debug('LogNormal mu:{} sigma:{} a:{} s:{} scale:{}'.format(
        #     mu, sigma, a, s, scale))

        # zeta = np.sqrt(np.log(1 + (sigma / mu) ** 2)) # shape factor
        # lambda_value = np.log(mu) - 0.5 * zeta ** 2   # expected value of ln x
        # logger.debug('LogNormal mu:{} sigma:{} zeta:{} lambda:{}'.format(
        #     mu, sigma, zeta, lambda_value))

        # self.rv = sps.lognorm(
        #     s=sigma,
        #     loc=mu)

        # self.rv = sps.lognorm(
        # 	s=s,
        #     loc=0,
        #     scale=scale)

    def get_random_value(self):
        val = np.random.lognormal(self.mu, self.sigma)
        return val

    def quantile(self, p):
        Q = math.sqrt(2.0 * self.sigma * self.sigma)
        Q *= erfinv(2.0 * p - 1.0)
        Q += self.mu
        Q = math.exp(Q)
        return Q

    def cdf(self, x):
        # Avoid SciPy LogNormal implementation
        #  since it inteprets mu differently
        if (x < MIN_X):
            return 0.0

        CDF = 0.5 * (1.0 + math.erf((np.log(x) - self.mu) / (self.sigma * SQRT2)))
        return CDF;

    def pdf(self, x):
        # Avoid SciPy LogNormal implementation
        #  since it inteprets mu differently
        if (x < MIN_X):
            return 0.0

        PDF = 1.0 / (x * self.sigma * SQRT2PI)
        PDF *= math.exp(-1.0 * ((np.log(x) - self.mu) ** 2) / (2.0 * self.sigma * self.sigma))
        return PDF

    def get_x_min(self):
        return 0

    def get_x_max(self):
        return math.inf



import logging
import numpy as np
import scipy.stats as sps
import math
from typing import List, Literal

from abc import abstractmethod

from .distribution import Distribution

logger = logging.getLogger(__name__)

class Analytic(Distribution):
    _rv: None = None
    _rv_cache: None = None
    _rv_index: int = 0

    def pdf(self, x:float) -> float:
        return self._rv.pdf(x)

    def cdf(self, x:float) -> float:
        return self._rv.cdf(x)

    def quantile(self, p:float) -> float:
        return self._rv.ppf(p)

    def get_random_value(self, size=1) -> float:
        if self._rv_cache is None:
            self._rv_cache = self.rv.rvs(10001)

        val = self._rv_cache[self.rv_index % 10001]
        self._rv_index += 1



import math
import sys
import numpy as np
from .distribution import Distribution

import marshmallow as m
import logging
logger = logging.getLogger(__name__)


class Analytic(Distribution):
    class ConfigSchema(Distribution.ConfigSchema):
        class Meta:
            include = {
                # 'min': m.fields.Float(missing=-sys.float_info.max),
                # 'max': m.fields.Float(missing=sys.float_info.max),
            }

    def load_config(self, config):
        super().load_config(config)
        # self.min_rv = config['min']
        # self.max_rv = config['max']

    def cache_init(self):
        self.x_space = None
        self.rv_cache = None
        self.rv_index = 0

    def get_random_value(self):
        if self.rv_cache is None:
            self.rv_cache = self.rv.rvs(10001)

        val = self.rv_cache[self.rv_index % 10001]
        self.rv_index += 1

        # Test comment
        # val = self.rv.rvs(size=1)[0]
        # if val > self.max_rv:
        #     return self.max_rv
        # if val < self.min_rv:
        #     return self.min_rv

        return val

    def quantile(self, p):
        return self.rv.ppf(p)

    def cdf(self, x):
        return self.rv.cdf(x)

    def pdf(self, x):
        return self.rv.pdf(x)

    def get_x_min(self):
        return -math.inf

    def get_x_max(self):
        return math.inf

    def get_percentile(self, p):
        return self.quantile(p)

    def get_x_space(self):
        if self.x_space is None:
            self.x_space = np.linspace(self.rv.ppf(0.01), self.rv.ppf(0.99), 100)
        return self.x_space


from keyrock_hcs import hcs

from .constant import Constant
from .uniform import Uniform
from .logistic import Logistic
from .normal import Normal
from .skew_normal import SkewNormal
from .log_normal import LogNormal
from .beta import Beta
from .gamma import Gamma

type_map = {
    'constant': Constant,
    'uniform': Uniform,
    'logistic': Logistic,
    'normal': Normal,
    'skew_normal': SkewNormal,
    'log_normal': LogNormal,
    'beta': Beta,
    'gamma': Gamma,
}

factory = hcs.TypeConfigFactory(type_map)



import scipy.stats as sps

from .analytic import Analytic

import marshmallow as m
import logging
logger = logging.getLogger(__name__)


class Uniform(Analytic):
    class ConfigSchema(Analytic.ConfigSchema):
        class Meta:
            include = {
                'min': m.fields.Float(missing=0),
                'max': m.fields.Float(missing=1),
            }

    def load_config(self, config):
        super().load_config(config)
        self.min_rv = config['min']
        self.max_rv = config['max']
        scale = self.max_rv - self.min_rv
        self.rv = sps.uniform(loc=self.min_rv, scale=scale)

    def get_x_min(self):
        return self.min_rv

    def get_x_max(self):
        return self.max_rv



import logging
import scipy.stats as sps
import math
from typing import Literal

from .analytic import Analytic

logger = logging.getLogger(__name__)

class Beta(Analytic):
    t: Literal['beta'] = 'beta'
    a: float = 1
    b: float = 1

    def __post_init__(self):
        self._rv = sps.beta(
                a=self.a,
                b=self.b,
                loc=0,
                scale=1
            )

    @property
    def x_min(self):
        return 0.0

    @property
    def x_max(self):
        return 1.0



import logging
import scipy.stats as sps
import math
from typing import Literal

from .analytic import Analytic

logger = logging.getLogger(__name__)

class Gamma(Analytic):
    t: Literal['gamma'] = 'gamma'
    a: float = 1
    rate: float = 1
    loc: float = 0

    def __post_init__(self):
        scale = 1.0 / self.rate
        self._rv = sps.gamma(
            a=self.a,
            loc=self.loc,
            scale=scale)

    @property
    def x_min(self):
        return 0.0

    @property
    def x_max(self):
        return math.inf



import numpy as np
import scipy.stats as sps

from .analytic import Analytic

import marshmallow as m
import logging
logger = logging.getLogger(__name__)


class Gamma(Analytic):
    class ConfigSchema(Analytic.ConfigSchema):
        class Meta:
            include = {
                'a': m.fields.Float(missing=1),
                'rate': m.fields.Float(missing=1),
                'loc': m.fields.Float(missing=0),
            }

    def load_config(self, config):
        super().load_config(config)
        scale = 1.0 / config['rate']
        self.rv = sps.gamma(
            a=config['a'],
            loc=config['loc'],
            scale=scale)



import logging
import scipy.stats as sps
import math
from typing import Literal

from .analytic import Analytic

logger = logging.getLogger(__name__)

class Uniform(Analytic):
    t: Literal['uniform'] = 'uniform'
    a: float = 0
    b: float = 1

    def __post_init__(self):
        scale = self.b - self.a
        self._rv = sps.uniform(
                loc=self.a,
                scale=scale
            )

    @property
    def x_min(self):
        return self.a

    @property
    def x_max(self):
        return self.b



import logging
import math
import numpy as np
from scipy.special import erfinv
from typing import Literal

from .analytic import Distribution

logger = logging.getLogger(__name__)

SQRT2 = math.sqrt(2)
SQRT2PI = math.sqrt(2 * math.pi)
MIN_X = 0.000000001

class LogNormal(Distribution):
    t: Literal['log_normal'] = 'log_normal'
    loc: float = 0
    scale: float = 1

    def __post_init__(self):
        self.mu = self.loc
        self.sigma = self.scale

    @property
    def x_min(self):
        return 0.0

    @property
    def x_max(self):
        return math.inf

    def pdf(self, x: float) -> float:
        # Avoid SciPy LogNormal implementation
        #  since it inteprets mu differently
        if (x < MIN_X):
            return 0.0

        PDF = 1.0 / (x * self.sigma * SQRT2PI)
        PDF *= math.exp(-1.0 * ((np.log(x) - self.mu) ** 2) / (2.0 * self.sigma * self.sigma))
        return PDF

    def cdf(self, x: float) -> float:
        # Avoid SciPy LogNormal implementation
        #  since it inteprets mu differently
        if (x < MIN_X):
            return 0.0

        CDF = 0.5 * (1.0 + math.erf((np.log(x) - self.mu) / (self.sigma * SQRT2)))
        return CDF;

    def quantile(self, p: float) -> float:
        Q = math.sqrt(2.0 * self.sigma * self.sigma)
        Q *= erfinv(2.0 * p - 1.0)
        Q += self.mu
        Q = math.exp(Q)
        return Q

    def get_random_value(self, size=1) -> float:
        return np.random.lognormal(self.mu, self.sigma)



import logging
import scipy.stats as sps
import math
from typing import Literal

from .analytic import Analytic

logger = logging.getLogger(__name__)

class Logistic(Analytic):
    t: Literal['logistic'] = 'logistic'
    loc: float = 0
    scale: float = 1

    def __post_init__(self):
        self._rv = sps.logistic(
                self.loc,
                self.scale
            )



import logging
import numpy as np
import scipy.stats as sps
import math
from typing import List, Literal

from abc import abstractmethod
from keyrock_model.model import KrBaseModel

logger = logging.getLogger(__name__)

class Distribution(KrBaseModel):
    t: Literal['base'] = 'base'
    _x_space: None = None

    @property
    def x_min(self):
        return -math.inf

    @property
    def x_max(self):
        return math.inf

    @abstractmethod
    def pdf(self, x: float) -> float:
        raise NotImplementedError()

    @abstractmethod
    def cdf(self, x: float) -> float:
        raise NotImplementedError()

    @abstractmethod
    def quantile(self, p: float) -> float:
        raise NotImplementedError()

    def get_random_value(self, size=1) -> float:
        raise NotImplementedError()

    ### Helpers ###

    def get_percentile(self, p):
        return self.quantile(p)

    def get_x_space(self):
        if self._x_space is None:
            self._x_space = np.linspace(self.quantile(0.01), self.quantile(0.99), 100)
        return self._x_space



import numpy as np
import scipy.stats as sps

from .analytic import Analytic

import marshmallow as m
import logging
logger = logging.getLogger(__name__)


class Normal(Analytic):
    class ConfigSchema(Analytic.ConfigSchema):
        class Meta:
            include = {
                'loc': m.fields.Float(missing=0),
                'scale': m.fields.Float(missing=1),
            }

    def load_config(self, config):
        super().load_config(config)
        self.rv = sps.norm(
        	loc=config['loc'],
        	scale=config['scale'])



import marshmallow as m
import numpy as np
import scipy.stats as sps
import math

import logging
logger = logging.getLogger(__name__)

from keyrock_hcs import hcs


class Distribution(hcs.HCSObject):
    class ConfigSchema(hcs.HCSObject.ConfigSchema):
        class Meta:
            pass

    def  __init__(self, config=None):
        super().__init__(config)
        self.cache_init()

    def cache_init(self):
        pass

    def get_x_min(self):
        return -math.inf

    def get_x_max(self):
        return math.inf

    def get_percentile(self, p):
        raise NotImplementedError()

    def get_x_space(self):
        raise NotImplementedError()

    def get_random_value(self, size=1):
        raise NotImplementedError()

    def quantile(self, p):
        raise NotImplementedError()

    def cdf(self, x):
        raise NotImplementedError()

    def pdf(self, x):
        raise NotImplementedError()


import logging
import scipy.stats as sps
import math
from typing import Literal

from .analytic import Analytic

logger = logging.getLogger(__name__)

class SkewNormal(Analytic):
    t: Literal['skew_normal'] = 'skew_normal'
    loc: float = 0
    scale: float = 1
    a: float = 0

    def __post_init__(self):
        self._rv = sps.skewnorm(
                loc=self.loc,
                scale=self.scale,
                a=self.a
            )



from .distribution import Distribution

import marshmallow as m
import logging
logger = logging.getLogger(__name__)


class Constant(Distribution):
    class ConfigSchema(Distribution.ConfigSchema):
        loc = m.fields.Float(missing=0)

    def load_config(self, config):
        super().load_config(config)
        self.v = config['loc']

    def get_x_min(self):
        return self.v

    def get_x_max(self):
        return self.v

    def get_percentile(self, p):
        return self.v

    def get_random_value(self, size=1):
        return self.v

    def cdf(self, x):
        return self.v * x

    def pdf(self, x):
        return self.v


import logging
from typing import Literal

from .analytic import Distribution

logger = logging.getLogger(__name__)

class Constant(Distribution):
    t: Literal['constant'] = 'constant'
    loc: float = 0

    @property
    def x_min(self):
        return self.loc

    @property
    def x_max(self):
        return self.loc

    def pdf(self, x: float) -> float:
        return self.loc

    def cdf(self, x: float) -> float:
        return self.loc * x

    def quantile(self, p: float) -> float:
        return self.loc

    def get_random_value(self, size=1) -> float:
        return self.loc
